class RectI:
    def __init__(self, il, it, ir, ib):
        """
        :type il: int
        :type it: int
        :type ir: int
        :type ib: int
        """
        if int != type(il):
            raise ValueError

        self.il = il
        self.it = it
        self.ir = ir
        self.ib = ib

        self.iw = ir - il
        self.ih = ib - it

    def __contains__(self, point_xy):
        """
        :type point_xy: (int, int)
        :rtype: bool
        """
        same_x = self.il < point_xy[0] < self.ir
        same_y = self.it < point_xy[1] < self.ib
        return same_x and same_y

    @staticmethod
    def create(rect_f, image_w, image_h):
        """
        :type rect_f: RectF
        :type image_w: int
        :type image_h: int
        :rtype: RectI
        """
        return RectI(int(rect_f.fl * image_w), int(rect_f.ft * image_h),
                     int(rect_f.fr * image_w), int(rect_f.fb * image_h))

    @property
    def center(self):
        """
        :rtype: (int, int)
        """
        return int((self.il + self.ir) / 2), int((self.it + self.ib) / 2)

    def serialize(self):
        """
        :rtype: list of int
        """
        return [self.il, self.it, self.ir, self.ib]


class RectF:
    def __init__(self, fl, ft, fr, fb):
        """
        :type fl: float
        :type ft: float
        :type fr: float
        :type fb: float
        """
        self.fl = fl
        self.ft = ft
        self.fr = fr
        self.fb = fb
        self.fw = fr - fl
        self.fh = fb - ft

    @staticmethod
    def create(center_l, center_t, width, height):
        """
        :type center_l: int | float
        :type center_t: int | float
        :type width: int | float
        :type height: int | float
        :rtype: RectF
        """
        center_fl, center_ft = float(center_l), float(center_t)
        half_fw, half_fh = width / 2.0, height / 2.0
        return RectF(
            center_fl - half_fw, center_ft - half_fh,
            center_fl + half_fw, center_ft + half_fh)

    @property
    def center(self):
        """
        :rtype: (float, float)
        """
        return (self.fl + self.fr) / 2.0, (self.ft + self.fb) / 2.0

    @property
    def area(self):
        """
        :rtype: float
        """
        return self.fw * self.fh

    def get_visible_100(self, image_w=1.0, image_h=1.0):
        """
        :type image_w: int | float
        :type image_h: int | float
        :rtype: bool
        """
        return self.fl > 0.0 and self.fr < image_w and self.ft > 0.0 and self.fb < image_h

    def get_visible(self, image_w=1.0, image_h=1.0):
        """
        :type image_w: int | float
        :type image_h: int | float
        :returns: value from 0.0% to 100.0%
        :rtype: float
        """
        self_area = self.area
        if self_area > 0.0:
            self_visible = RectF(self.fl, self.ft, self.fr, self.fb)
            self_visible.strip_invisible(image_w, image_h)
            return 100.0 * self_visible.area / self_area

        else:
            return 0.0

    def strip_invisible(self, image_w=1.0, image_h=1.0):
        """
        :type image_w: int | float
        :type image_h: int | float
        :raise ValueError:
        """
        def norm(max_coord, coord):
            """
            :type max_coord: int | float
            :type coord: float
            :rtype: float
            """
            if coord < 0.0:
                return 0.0

            elif coord > max_coord:
                return float(max_coord)

            else:
                return coord

        self.fl = norm(image_w, self.fl)
        self.ft = norm(image_h, self.ft)
        self.fr = norm(image_w, self.fr)
        self.fb = norm(image_h, self.fb)

        self.fw = self.fr - self.fl
        self.fh = self.fb - self.ft

    def make_visible(self, image_w=1.0, image_h=1.0):
        """
        :type image_w: int | float
        :type image_h: int | float
        :raise ValueError:
        """
        if self.fw > image_w or self.fh > image_h:
            raise ValueError

        # move self to a visible part of the image
        next_fl = self.fl
        if self.fl < 0.0:
            next_fl = 0.0

        elif self.fl + self.fw > image_w:
            next_fl = image_w - self.fw

        next_ft = self.ft
        if self.ft < 0.0:
            next_ft = 0.0

        elif self.ft + self.fh > image_h:
            next_ft = image_h - self.fh

        # fix rounding errors if any
        next_fw, next_fh = self.fw, self.fh

        next_fr = next_fl + next_fw
        if next_fr >= image_w:
            next_fw -= 1.0

            next_fr = next_fl + next_fw
            if next_fr >= image_w:
                raise ValueError

        next_fb = next_ft + next_fh
        if next_fb >= image_h:
            next_fh -= 1.0

            next_fb = next_ft + next_fh
            if next_fb >= image_h:
                raise ValueError

        # actualize self coordinates
        self.fl = next_fl
        self.ft = next_ft
        self.fr = next_fr
        self.fb = next_fb

        self.fw = self.fr - self.fl
        self.fh = self.fb - self.ft

    def scale_max(self, min_w=0.0, min_h=0.0, max_w=1.0, max_h=1.0):
        """
        :type min_w: int | float
        :type min_h: int | float
        :type max_w: int | float
        :type max_h: int | float
        :raise ValueError:
        """
        assert self.fh > 0.0
        wh_ratio = self.fw / self.fh

        def get_new_wh():
            """
            :rtype: (float, float)
            :raises ValueError:
            """
            if self.fw <= max_w and self.fh <= max_h:
                return self.fw, self.fh

            # strip extra h-pixels from self height (and keep aspect-ratio)
            if self.fh > max_h:
                next_fw = max_h * wh_ratio
                if min_w < next_fw <= max_w:
                    return next_fw, max_h

            # strip extra w-pixels from self width (and keep aspect-ratio)
            if self.fw > max_w:
                next_fh = max_w / wh_ratio
                if min_h < next_fh <= max_h:
                    return max_w, next_fh

            raise ValueError

        self.fw, self.fh = get_new_wh()

        half_fw, half_fh = self.fw / 2.0, self.fh / 2.0
        center_fl, center_ft = self.center

        self.fl = center_fl - half_fw
        self.ft = center_ft - half_fh
        self.fr = center_fl + half_fw
        self.fb = center_ft + half_fh


class QuadI:
    def __init__(self, x0, y0, x1, y1, x2, y2, x3, y3):
        """
        :type x0: int
        :type y0: int
        :type x1: int
        :type y1: int
        :type x2: int
        :type y2: int
        :type x3: int
        :type y3: int
        """

        self.x0 = x0
        self.y0 = y0

        self.x1 = x1
        self.y1 = y1

        self.x2 = x2
        self.y2 = y2

        self.x3 = x3
        self.y3 = y3

    def __contains__(self, point_xy):
        """
        :type point_xy: (int, int)
        :rtype: bool
        """
        min_x = min(self.x0, self.x3)
        max_x = max(self.x1, self.x2)
        min_y = min(self.y0, self.y1)
        max_y = max(self.y2, self.y3)

        same_x = min_x < point_xy[0] < max_x
        same_y = min_y < point_xy[1] < max_y
        return same_x and same_y

    def map(self, source_im_w, source_im_h, target_im_w, target_im_h, margin_tb=0, margin_lr=0):
        """
        :type source_im_w: int
        :type source_im_h: int
        :type target_im_w: int
        :type target_im_h: int
        :type margin_tb: int
        :type margin_lr: int

        :rtype: (int, int, int, int, int, int, int, int)
        """

        x0 = int(float(self.x0) / source_im_w * target_im_w - margin_lr)
        y0 = int(float(self.y0) / source_im_h * target_im_h - margin_tb)

        x1 = int(float(self.x1) / source_im_w * target_im_w + margin_lr)
        y1 = int(float(self.y1) / source_im_h * target_im_h - margin_tb)

        x2 = int(float(self.x2) / source_im_w * target_im_w + margin_lr)
        y2 = int(float(self.y2) / source_im_h * target_im_h + margin_tb)

        x3 = int(float(self.x3) / source_im_w * target_im_w - margin_lr)
        y3 = int(float(self.y3) / source_im_h * target_im_h + margin_tb)

        return x0, y0, x1, y1, x2, y2, x3, y3

    def get_abs_bbox(self):
        """
        :rtype: (int, int, int, int)
        """
        il = min(self.x0, self.x3)
        ir = max(self.x1, self.x2)
        it = min(self.y0, self.y1)
        ib = max(self.y2, self.y3)

        return il, it, ir, ib

    def get_rel_bbox(self, source_im_w, source_im_h):
        """
        :type source_im_w: int
        :type source_im_h: int
        :rtype: (float, float, float, float)
        """
        fl = float(min(self.x0, self.x3)) / source_im_w
        fr = float(max(self.x1, self.x2)) / source_im_w
        ft = float(min(self.y0, self.y1)) / source_im_h
        fb = float(max(self.y2, self.y3)) / source_im_h

        return fl, ft, fr, fb

    def map_bbox_wh(self, source_im_w, source_im_h, target_im_w, target_im_h):
        """
        :type source_im_w: int
        :type source_im_h: int
        :type target_im_w: int
        :type target_im_h: int
        :rtype: (int, int)
        """
        bbox_il = min(self.x0, self.x3)
        bbox_ir = max(self.x1, self.x2)
        bbox_it = min(self.y0, self.y1)
        bbox_ib = max(self.y2, self.y3)

        target_bbox_fw = float(bbox_ir - bbox_il) / source_im_w * target_im_w
        target_bbox_fh = float(bbox_ib - bbox_it) / source_im_h * target_im_h
        return int(target_bbox_fw), int(target_bbox_fh)

    def map_bbox_ref(self, source_im_w, source_im_h, target_im_w, target_im_h):
        """
        :type source_im_w: int
        :type source_im_h: int
        :type target_im_w: int
        :type target_im_h: int
        :rtype: (int, int)
        """
        target_bbox_il = float(min(self.x0, self.x3)) / source_im_w * target_im_w
        target_bbox_ir = float(max(self.x1, self.x2)) / source_im_w * target_im_w
        target_bbox_fb = float(max(self.y2, self.y3)) / source_im_h * target_im_h

        return int(0.5 * (target_bbox_il + target_bbox_ir)), int(target_bbox_fb)

    @property
    def center(self):
        """
        :rtype: (int, int)
        """
        return RectI(*self.get_abs_bbox()).center

    @property
    def wh(self):
        """
        :rtype: (int, int)
        """
        bbox = RectI(*self.get_abs_bbox())
        return bbox.iw, bbox.ih

    def serialize(self):
        """
        :rtype: list[int]
        """
        return [self.x0, self.y0, self.x1, self.y1,
                self.x2, self.y2, self.x3, self.y3]

    @staticmethod
    def unpack(jnode):
        """
        :type jnode: list[int]
        :rtype: QuadI
        """
        if len(jnode) != 8:
            raise RuntimeError

        return QuadI(*jnode)


def mk_plate_bbox(plate_rect, image_iw, image_ih):
    """
    :param plate_rect: relative coords (visible values from 0.0 to 1.0)
    :type plate_rect: RectF
    :param image_iw: width, px
    :type image_iw: int
    :param image_ih: height, px
    :type image_ih: int
    :rtype: RectF
    """
    fl = image_iw * plate_rect.fl
    ft = image_ih * plate_rect.ft
    fr = image_iw * plate_rect.fr
    fb = image_ih * plate_rect.fb

    # tweak plate rect for better view
    sym_margin_cm, sym_height_cm = 1.5, 8.0
    delta = (fb - ft) * sym_margin_cm / sym_height_cm
    return RectF(
        fl - 3 * delta,
        ft - 1 * delta,
        fr + 2 * delta,
        fb + 1 * delta)


def mk_plate_bbox2(plate_rect, image_iw, image_ih):
    """
    :param plate_rect: relative coords (visible values from 0.0 to 1.0)
    :type plate_rect: RectF
    :param image_iw: width, px
    :type image_iw: int
    :param image_ih: height, px
    :type image_ih: int
    :rtype: RectF
    """
    fl = image_iw * plate_rect.fl
    ft = image_ih * plate_rect.ft
    fr = image_iw * plate_rect.fr
    fb = image_ih * plate_rect.fb

    return RectF(fl, ft, fr, fb)


def _upscale_plate_wh(plate_bbox, wh_ratio, target_w):
    """
    :param plate_bbox: absolute coords
    :type plate_bbox: RectF
    :type wh_ratio: float
    :type target_w: float

    Vehicle's make and model would be recognizable in case of wh_ratio 4.0/3.0
    :rtype: (float, float)
    """
    # define the width of a vehicle as the width of the lane (typ. 2.75m)
    if plate_bbox.fw / plate_bbox.fh > 2.0:
        # single-line plates with w/h-ratio more then 2:1 (typ. 5:1)
        # suppose the average plate's height that was defined by ANPR is around 10cm
        # suppose the average plate's width that was defined by ANPR is around 50cm
        car_w_height = (target_w / 0.10) * plate_bbox.fh
        car_w_width = (target_w / 0.50) * plate_bbox.fw
        car_w = (car_w_height + car_w_width) / 2

    else:
        # double-line plates with w/h-ratio 2:1 or less
        # suppose the average plate's height that was defined by ANPR is around 17cm
        # suppose the average plate's width that was defined by ANPR is around 22cm
        car_w_height = (target_w / 0.17) * plate_bbox.fh
        car_w_width = (target_w / 0.22) * plate_bbox.fw
        car_w = (car_w_height + car_w_width) / 2

    return car_w, car_w / wh_ratio


def mk_vehicle_bbox(plate_bbox, scene_wh, wh_ratio=4.0/3.0):
    """
    :param plate_bbox: absolute coords
    :type plate_bbox: RectF
    :type scene_wh: (int, int)
    :type wh_ratio: float

    :return: absolute coords
    :rtype: RectF
    :raise ValueError:
    """
    center_fl, center_ft = plate_bbox.center
    width, height = _upscale_plate_wh(plate_bbox, wh_ratio, 2.75)
    bbox = RectF.create(center_fl, center_ft - height / 5.0, width, height)  # move bbox 20% up from plate center
    bbox.scale_max(plate_bbox.fw, plate_bbox.fh, *scene_wh)
    bbox.make_visible(*scene_wh)
    return bbox


def mk_mmr_bbox(plate_bbox, scene_wh, min_visible_factor):
    """
    :param plate_bbox: absolute coords
    :type plate_bbox: RectF
    :type scene_wh: (int, int)
    :type min_visible_factor: float

    :return: absolute coords
    :rtype: RectF
    :raise ValueError:
    """
    center_fl, center_ft = plate_bbox.center
    width, height = _upscale_plate_wh(plate_bbox, 1, 2.0)
    bbox = RectF.create(center_fl, center_ft - height / 4.0, width, height)  # move bbox 25% up from plate center
    if bbox.get_visible(*scene_wh) < min_visible_factor:
        raise ValueError

    bbox.scale_max(plate_bbox.fw, plate_bbox.fh, *scene_wh)
    bbox.make_visible(*scene_wh)
    return bbox
